# ESA (C) 2000-2020
# 
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS.  If not, see <http://www.gnu.org/licenses/>.

# startsas.py


# Standard library imports
import os, sys, subprocess, shutil, glob

# Third party imports
# (se below for astroquery)

# Local application imports
from .version import VERSION, SAS_RELEASE, SAS_AKA
from pysas.wrapper import Wrapper as wrap


__version__ = f'startsas (startsas-{VERSION}) [{SAS_RELEASE}-{SAS_AKA}]' 

# The Python code for startsas should go inside run

# iparsdic is a dictionary with all the task parameters, where their
# respective values are either those entered from the command line 
# or those taken from par file defaults.


def run(iparsdic):
    '''
    startsas could be used in a Python session to start working with SAS.

    It makes the following assumptions for the terminal being used:

        1. It has the right Python base environment:
            - Python >= 3.8
            - astropy, numpy, matplotlib, requests, pyds9, PyQt5,
              beautifultable, astroquery, Jupyter Notebooks.

        2. Heasoft has been initialised, so that LHEASOFT is set.

        3. SAS has been initialised, so that SAS_DIR, SAS_PATH and some
        more SAS specific env. variables are already set and point to a
        valid SAS installation.

        4. SAS_CCFPATH is set pointing to the CCF repository.

    '''

    # Checking LHEASOFT, SAS_DIR and SAS_CCFPATH

    lheasoft = os.environ.get('LHEASOFT')
    if not lheasoft:
        raise Exception('LHEASOFT is not set. Please initialise HEASOFT')

    sasdir = os.environ.get('SAS_DIR')
    if not sasdir:
        raise Exception('SAS_DIR is not defined. Please initialise SAS')

    sasccfpath = os.environ.get('SAS_CCFPATH')
    if not sasccfpath:
        raise Exception('SAS_CCFPATH not set. Please define it')

    # Where are we?
    where = os.getcwd()
    print(f'''

    Starting SAS session
    
    Working directory = {where}
    
    ''')

    # Identify the download level
    level = iparsdic['level']
    if level != 'ODF' and level != 'PPS':
        raise Exception('ODF request level is undefined!')


    # Processing odfid
    if iparsdic['odfid'] and level == 'ODF':
        
        if iparsdic['sas_ccf'] or iparsdic['sas_ccf']:
            raise Exception('Parameter odfid is icompatible with sas_ccf and sas_odf')

        odfid = iparsdic['odfid']
        print('Requesting odfid  = {} to XMM-Newton Science Archive\n'.format(iparsdic['odfid']))
        
        # Download the odfid from XMM-Newton, using astroquery

        from astroquery.esa.xmm_newton import XMMNewton
        print(f'\nDownloading {odfid}, level {level}. Please wait ...\n')
        XMMNewton.download_data(odfid, level=level)

        tarfile = odfid + '.tar.gz'
        
        # Check that the tar.gz file has been downloaded
        try:
            os.path.exists(tarfile)
        except FileExistsError:
            print(f'File {tarfile} is not present. Not downloaded?')
            sys.exit(1)
        
        # Creates subdirectory odfid to move and unpack the odfid.tar.gz file
        print('\n\nWarning: Any *.cif and *SUM.SAS files existing here will be removed!')
        if os.path.exists(os.path.join(where, odfid)):
            print(f'\nRemoving directory {odfid} ...')
            shutil.rmtree(os.path.join(where, odfid))

        print(f'\nCreating directory {odfid} ...')
        os.mkdir(odfid)
        
        # Moves odfid.tar.gz file to odfid
        shutil.move(tarfile, odfid)
        
        # Changes dir to odfid
        os.chdir(odfid)
        
        # Untars the odfid.tar.gz file
        cmd = ['tar', 'zxf', tarfile]
        print(f'\nUnpacking {tarfile} ...\n')
        rc = subprocess.run(cmd)
        if rc.returncode != 0:
            raise Exception('tar file extraction failed')
        
        # Obtains the name of the file with ext TAR
        TARFILE = glob.glob('*.TAR')
        cmd = ['tar', 'xf', TARFILE[0]]
        # Untars the TAR file
        print(f'Unpacking {TARFILE[0]} ...')
        rc = subprocess.run(cmd)
        
        # Checks that the MANIFEST file is there
        MANIFEST = glob.glob('MANIFEST*')
        try:
            os.path.exists(MANIFEST[0])
        except FileExistsError:
            print(f'File {MANIFEST[0]} not present. Please check ODF!')
            sys.exit(1)

        # Here the ODF is fully untarred below odfid subdirectory
        # Now we start preparing the SAS_ODF and SAS_CCF
        print(f'\nSetting SAS_ODF = {os.getcwd()}')
        os.environ['SAS_ODF'] = os.getcwd()

        # Change back to where
        os.chdir(where)

        # Run cifbuild
        if iparsdic['cifbuild_opts']:
            cifbuild_opts = iparsdic['cifbuild_opts']
            cmd = ['cifbuild', cifbuild_opts]
            print(f'\nRunning cifbuild with {cifbuild_opts} ...')
        else:
            cmd = ['cifbuild']
            print(f'\nRunning cifbuild...')
        
        rc = subprocess.run(cmd)
        if rc.returncode != 0:
            raise Exception('cifbuild failed to complete')
        
        # Check whether ccf.cif is produced or not
        ccfcif = glob.glob('ccf.cif')
        try:
            os.path.exists(ccfcif[0])
        except FileExistsError:
            print('ccf.cif file is not produced')
            sys.exit(1)
        
        # Sets SAS_CCF variable
        fullccfcif = os.path.join(where, 'ccf.cif')
        print(f'\nSetting SAS_CCF = {fullccfcif}')
        os.environ['SAS_CCF'] = fullccfcif

        # Now run odfingest
        if iparsdic['odfingest_opts']:
            odfingest_opts = iparsdic['odfingest_opts']
            cmd = ['odfingest', odfingest_opts]
            print(f'\nRunning odfingest with {odfiingest_opts} ...')
        else:
            cmd = ['odfingest']
            print('\nRunning odfingest...')
        
        rc = subprocess.run(cmd)
        if rc.returncode != 0:
            raise Exception('odfingest failed to complete.')

        # Check whether the SUM.SAS has been produced or not
        sumsas = glob.glob('*SUM.SAS')
        try:
            os.path.exists(sumsas[0])
        except FileExistsError:
            print('SUM.SAS files is not produced')
            sys.exit(1)
        
        # Set the SAS_ODF to the SUM.SAS file
        fullsumsas = os.path.join(where, sumsas[0])
        os.environ['SAS_ODF'] = fullsumsas
        print(f'\nSetting SAS_ODF = {fullsumsas}')

        # sasodf is the dirname of fullsumsas + odfid. It will be used below.
        sasodf = os.path.join(os.path.dirname(fullsumsas), odfid)

        # Check that the SUM.SAS file has the right PATH keyword
        with open(fullsumsas) as inf:
            lines = inf.readlines()
        for line in lines:
            if 'PATH' in line:
                key, path = line.split()
                if path != sasodf:
                    raise Exception(f'SAS summary file PATH mismatchs {sasodf}')
                else:
                    print(f'\nWarning: Summary file PATH keyword matches {sasodf}')

        print(f'''\n\n
        SAS_CCF = {fullccfcif}
        SAS_ODF = {fullsumsas}
        \n''')

    # Process odfid with level=PPS
    elif iparsdic['odfid'] and level == 'PPS':
        
        if iparsdic['sas_ccf'] or iparsdic['sas_ccf']:
            raise Exception('Parameter odfid is icompatible with sas_ccf and sas_odf')

        odfid = iparsdic['odfid']
        print('Requesting odfid  = {} to XMM-Newton Science Archive\n'.format(iparsdic['odfid']))

        # Download the odfid from XMM-Newton, using astroquery

        from astroquery.esa.xmm_newton import XMMNewton
        print(f'\nDownloading {odfid}, level {level}. Please wait ...\n')
        XMMNewton.download_data(odfid, level=level)

        tarfile = odfid + '.tar'

        # Check that the tar file has been downloaded
        try:
            os.path.exists(tarfile)
        except FileExistsError:
            print(f'File {tarfile} is not present. Not downloaded?')
            sys.exit(1)

        # Creates subdirectory odfid to move and unpack the odfid.tar file
        if os.path.exists(os.path.join(where, odfid)):
            print(f'\nRemoving directory {odfid} ...')
            shutil.rmtree(os.path.join(where, odfid))


        # Untars the odfid.tar.gz file
        cmd = ['tar', 'xf', tarfile]
        print(f'\nUnpacking {tarfile} ...\n')
        rc = subprocess.run(cmd)
        if rc.returncode != 0:
            raise Exception('tar file extraction failed')

        ppsdir = os.path.join(where, odfid, 'pps')
        ppssumhtml = 'P' + odfid + 'OBX000SUMMAR0000.HTM'
        ppssumhtmlfull = os.path.join(ppsdir, ppssumhtml)
        ppssumhtmllink = 'file://' + ppssumhtmlfull
        print(f'\nPPS products can be found in {ppsdir}\n\nLink to Observation Summary html: {ppssumhtmllink}')

    # Process sas_ccf and sas_odf parameters
    elif iparsdic['sas_ccf'] and not iparsdic['sas_odf']:
        raise Exception('sas_odf is not defined. You must define both, sas_ccf and sas_odf.')
    elif iparsdic['sas_odf'] and not iparsdic['sas_ccf']:
        raise Exception('sas_ccf is not defined. You must define both, sas_ccf and sas_odf.')
    elif iparsdic['sas_ccf'] and iparsdic['sas_odf']:
        if iparsdic['odfid']:
            raise Exception('Parameters sas_ccf and sas_odf are incompatible with parameter odfid')
        if 'ccf.cif' not in iparsdic['sas_ccf']:
            raise Exception('{} does not refer to a CIF file'.format(iparsdic['sas_ccf']))
        sasccf = iparsdic['sas_ccf']
        try:
            os.path.exists(sasccf)
        except FileExistsError:
            print(f'File {sasccf} is not found.')
            sys.exit(1)
        print('SAS_CCF = {}'.format(iparsdic['sas_ccf']))

        if 'SUM.SAS' not in iparsdic['sas_odf']:
            raise Exception('{} does not refer to a SAS SUM file'.format(iparsdic['sas_odf']))
        sasodf = iparsdic['sas_odf']
        try:
            os.path.exists(sasodf)
        except FileExistsError:
            print(f'File {sasodf} is not found.')
            sys.exit(1)
        print('SAS_ODF = {}'.format(iparsdic['sas_odf']))

        # Check that the SUM.SAS file PATH keyword points to a real ODF directory
        with open(sasodf) as inf:
            lines = inf.readlines()
        for line in lines:
            if 'PATH' in line:
                key, path = line.split()
                if not os.path.exists(path):
                    raise Exception(f'Summary file PATH {path} does not exist.')
                MANIFEST = glob.glob(os.path.join(path, 'MANIFEST*'))
                if not os.path.exists(MANIFEST[0]):
                    raise Exception(f'\nMissing {MANIFEST[0]} file in {path}. Missing ODF components?')
        
    else:
        print('''
                        Getting started with startsas
        
        After Heasoft and SAS initialisations, the quickest way of starting a
        working session with SAS is to run 
        
            startsas odfid=0122700101 
        
        where the value given to the 'odfid' parameter is the ODF ID of a 
        given XMM-Newton Observation. 

        The startsas program will understand you want to get such Observation
        from the XMM-Newton Science Archive. The download will be done by
        means of a special version of the Python module 'astroquery' prepared 
        to work with XMM-Newton data.
        
        By default data are obtained at level 'ODF' which includes only
        the data. The parameter named 'level' can be used to select the
        alternative level 'PPS', which will download the data and the 
        products obtained from the XMM-Newton Pipeline. 

        For level 'ODF', file <odfid>.tar.gz is downloaded to the 
        directory where you are and moved from there to subdirectory <odfid>.

        For level 'PPS', all Pipeleine products are placed in subdirectory
        <odfid>/pps. A link to the html including the Observation Summary
        (P<odfid>OBX000SUMMAR0000.HTM) is printed out.
        
        Instead of 'odfid', we can use parameters sas_ccf and sas_odf as

            startsas sas_ccf=<path>/ccf.cif sas_odf=<path>/*SUM.SAS

        The program understands you have already valid ccf.cif and SAS 
        summary files, in directory <path>, which you want to use. 
        
        Before using these values the program checks

        . The PATH keyword written inside the SAS summary file
        . The existence of the mandatory file MANIFEST.NNNNNN 
          (NNNNNN is the AMS extraction number) 
          
        to ensure there is actually a real ODF stored there.
        
        For the time being, the only check on ccf.cif is its existence. 
        ''')
